import re

LANGUAGE_RE = re.compile(
    "^(.*_(((([A-Za-z]{2,3}(-([A-Za-z]{3}(-[A-Za-z]{3}){0,2}))?)|[A-Za-z]{4}|[A-Za-z]{5,8})(-([A-Za-z]{4}))?(-([A-Za-z]{2}|[0-9]{3}))?(-([A-Za-z0-9]{5,8}|[0-9][A-Za-z0-9]{3}))*(-([0-9A-WY-Za-wy-z](-[A-Za-z0-9]{2,8})+))*(-(x(-[A-Za-z0-9]{1,8})+))?)|(x(-[A-Za-z0-9]{1,8})+)))$"
)


def schema_dict_fields_generator(schema_dict):
    if "properties" in schema_dict and isinstance(schema_dict["properties"], dict):
        for property_name, value in schema_dict["properties"].items():
            if "oneOf" in value:
                property_schema_dicts = value["oneOf"]
            else:
                property_schema_dicts = [value]
            for property_schema_dict in property_schema_dicts:
                if not isinstance(property_schema_dict, dict):
                    continue
                if "properties" in property_schema_dict:
                    for field in schema_dict_fields_generator(property_schema_dict):
                        yield f"/{property_name}{field}"
                elif "items" in property_schema_dict:
                    for field in schema_dict_fields_generator(
                        property_schema_dict["items"]
                    ):
                        yield f"/{property_name}{field}"
                yield f"/{property_name}"
    if "items" in schema_dict and isinstance(schema_dict["items"], dict):
        if "oneOf" in schema_dict["items"] and isinstance(
            schema_dict["items"]["oneOf"], list
        ):
            for oneOf in schema_dict["items"]["oneOf"]:
                for field in schema_dict_fields_generator(oneOf):
                    yield field


def get_additional_fields_info(json_data, schema_fields, fields_regex=False):
    fields_present = get_fields_present_with_examples(json_data)

    additional_fields = {}
    root_additional_fields = set()

    for field, field_info in fields_present.items():

        if field in schema_fields:
            continue
        if fields_regex and LANGUAGE_RE.search(field.split("/")[-1]):
            continue

        for root_additional_field in root_additional_fields:
            if field.startswith(root_additional_field):
                field_info["root_additional_field"] = False
                additional_fields[root_additional_field][
                    "additional_field_descendance"
                ][field] = field_info
                break
        else:
            field_info["root_additional_field"] = True
            field_info["additional_field_descendance"] = {}
            root_additional_fields.add(field)

        field_info["path"] = "/".join(field.split("/")[:-1])
        field_info["field_name"] = field.split("/")[-1]
        additional_fields[field] = field_info

    return additional_fields


def get_fields_present_with_examples(*args, **kwargs):
    counter = {}
    for key, value in fields_present_generator(*args, **kwargs):
        if key not in counter:
            counter[key] = {"count": 1, "examples": []}
        else:
            counter[key]["count"] += 1
        if len(counter[key]["examples"]) < 3:
            if not isinstance(value, (list, dict)):
                counter[key]["examples"].append(value)

    return counter


def fields_present_generator(json_data, prefix=""):
    if isinstance(json_data, dict):
        for key, value in json_data.items():
            new_key = f"{prefix}/{key}"
            yield new_key, value
            if isinstance(value, (dict, list)):
                yield from fields_present_generator(value, new_key)
    elif isinstance(json_data, list):
        for item in json_data:
            if isinstance(item, dict):
                yield from fields_present_generator(item, prefix)



import copy
from collections import defaultdict
from enum import Enum
from typing import Optional

from json_merge_patch import create_patch as json_diff_function

from libcove2.common import fields_present_generator


class GeoJSONAssumeFeatureType(Enum):
    NODE = "node"
    SPAN = "span"


class JSONToGeoJSONConverter:
    """Converts JSON data to GeoJSON."""

    def __init__(self):
        self._nodes_geojson_features: list = []
        self._spans_geojson_features: list = []

    def process_package(self, package_data: dict) -> None:
        """Process network package. Pass in data. Results are stored on object to get with other methods."""
        for network in package_data.get("networks", []):
            self._process_network(network)

    def _process_network(self, network_data: dict) -> None:
        nodes = network_data.pop("nodes", [])
        spans = network_data.pop("spans", [])
        phases = network_data.pop("phases", [])
        organisations = network_data.pop("organisations", [])

        # Dereference `phases.funders`
        for phase in phases:
            if "funders" in phase and isinstance(phase["funders"], list):
                phase["funders"] = [
                    self._dereference_object(organisation, organisations)
                    for organisation in phase["funders"]
                ]

        # Dereference `contracts.relatedPhases`
        if "contracts" in network_data and isinstance(network_data["contracts"], list):
            for contract in network_data["contracts"]:
                if "relatedPhases" in contract and isinstance(
                    contract["relatedPhases"], list
                ):
                    contract["relatedPhases"] = [
                        self._dereference_object(phase, phases)
                        for phase in contract["relatedPhases"]
                    ]

        # Convert nodes to features
        for node in nodes:
            self._nodes_geojson_features.append(
                self._convert_node_to_feature(node, network_data, organisations, phases)
            )

        # Convert spans to features
        for span in spans:
            self._spans_geojson_features.append(
                self._convert_span_to_feature(
                    span, network_data, organisations, phases, nodes
                )
            )

    def get_nodes_geojson(self) -> dict:
        """After processing, call to get nodes GeoJSON output."""
        return {"type": "FeatureCollection", "features": self._nodes_geojson_features}

    def get_spans_geojson(self) -> dict:
        """After processing, call to get spans GeoJSON output."""
        return {"type": "FeatureCollection", "features": self._spans_geojson_features}

    def get_meta_json(self) -> dict:
        """After processing, call to get meta information on the conversion."""
        out: dict = {
            "nodes_output_field_coverage": {},
            "spans_output_field_coverage": {},
        }
        # nodes field coverage
        for key, value in fields_present_generator(self.get_nodes_geojson()):
            if key not in out["nodes_output_field_coverage"]:
                out["nodes_output_field_coverage"][key] = {"count": 1}
            else:
                out["nodes_output_field_coverage"][key]["count"] += 1
        # spans field coverage
        for key, value in fields_present_generator(self.get_spans_geojson()):
            if key not in out["spans_output_field_coverage"]:
                out["spans_output_field_coverage"][key] = {"count": 1}
            else:
                out["spans_output_field_coverage"][key]["count"] += 1
        # Any geometries?
        out["any_spans_with_geometry"] = bool(
            [True for s in self._spans_geojson_features if s.get("geometry")]
        )
        out["any_nodes_with_geometry"] = bool(
            [True for n in self._nodes_geojson_features if n.get("geometry")]
        )
        # return
        return out

    def _dereference_object(self, ref, list):
        """
        Return from list the object referenced by ref. Otherwise, return ref.
        """

        if "id" in ref:
            for item in list:
                if isinstance(item, dict) and item.get("id") == ref["id"]:
                    return item

        return ref

    def _convert_node_to_feature(
        self,
        node_data: dict,
        reduced_network_data: dict,
        organisations: list,
        phases: list,
    ) -> dict:

        reduced_node_data = copy.deepcopy(node_data)

        feature = {
            "type": "Feature",
            "geometry": reduced_node_data.pop("location")
            if isinstance(reduced_node_data.get("location"), dict)
            else None,
        }

        # Dereference organisation references
        if isinstance(reduced_node_data.get("physicalInfrastructureProvider"), dict):
            reduced_node_data[
                "physicalInfrastructureProvider"
            ] = self._dereference_object(
                reduced_node_data["physicalInfrastructureProvider"], organisations
            )
        if isinstance(reduced_node_data.get("networkProviders"), list):
            reduced_node_data["networkProviders"] = [
                self._dereference_object(i, organisations)
                for i in reduced_node_data["networkProviders"]
                if isinstance(i, dict)
            ]

        # Dereference phase references
        if "phase" in reduced_node_data:
            reduced_node_data["phase"] = self._dereference_object(
                reduced_node_data["phase"], phases
            )

        feature["properties"] = reduced_node_data
        feature["properties"]["network"] = reduced_network_data
        feature["properties"]["featureType"] = "node"

        return feature

    def _convert_span_to_feature(
        self,
        span_data: dict,
        reduced_network_data: dict,
        organisations: list,
        phases: list,
        nodes: list,
    ) -> dict:

        reduced_span_data = copy.deepcopy(span_data)

        feature = {
            "type": "Feature",
            "geometry": reduced_span_data.pop("route")
            if isinstance(reduced_span_data.get("route"), dict)
            else None,
        }

        # Dereference organisation references
        if isinstance(reduced_span_data.get("physicalInfrastructureProvider"), dict):
            reduced_span_data[
                "physicalInfrastructureProvider"
            ] = self._dereference_object(
                reduced_span_data["physicalInfrastructureProvider"], organisations
            )
        if isinstance(reduced_span_data.get("networkProviders"), list):
            reduced_span_data["networkProviders"] = [
                self._dereference_object(i, organisations)
                for i in reduced_span_data["networkProviders"]
                if isinstance(i, dict)
            ]

        # Dereference phase references
        if "phase" in reduced_span_data:
            reduced_span_data["phase"] = self._dereference_object(
                reduced_span_data["phase"], phases
            )

        # Dereference endpoints
        for endpoint in ["start", "end"]:
            if endpoint in reduced_span_data:
                for node in nodes:
                    if "id" in node and node["id"] == reduced_span_data[endpoint]:
                        reduced_span_data[endpoint] = node

        feature["properties"] = reduced_span_data
        feature["properties"]["network"] = reduced_network_data
        feature["properties"]["featureType"] = "span"

        return feature


class GeoJSONToJSONConverter:
    """Converts GeoJSON data to JSON."""

    def __init__(self):
        self._networks: dict = {}
        self._inconsistent_phase_ids_by_network_id: defaultdict = defaultdict(set)
        self._inconsistent_organisation_ids_by_network_id: defaultdict = defaultdict(
            set
        )
        self._inconsistent_network_ids_seen: set = set()

    def process_data(
        self,
        data: dict,
        assumed_feature_type: GeoJSONAssumeFeatureType = GeoJSONAssumeFeatureType.NODE,
    ) -> None:
        """Process data. Results are stored on object to get with other methods.

        Can be called multiple times with as many GeoJSON files as needed.
        """
        # Network
        for geojson_feature in data.get("features", []):
            self._process_network(geojson_feature)

        # Nodes/Spans
        for geojson_feature in data.get("features", []):
            type = assumed_feature_type.value
            if isinstance(geojson_feature.get("properties"), dict) and isinstance(
                geojson_feature["properties"].get("featureType"), str
            ):
                type = geojson_feature["properties"]["featureType"]
            if type.lower() == "node":
                self._process_node(geojson_feature)
            elif type.lower() == "span":
                self._process_span(geojson_feature)
            else:
                # TODO log error
                pass

    def _process_network(self, geojson_feature_node_or_span: dict) -> None:
        if (
            "properties" in geojson_feature_node_or_span
            and "network" in geojson_feature_node_or_span["properties"]
        ):
            network = geojson_feature_node_or_span["properties"]["network"]
            if isinstance(network, dict):
                network_id = network.get("id")
                if isinstance(network_id, str) and network_id:

                    # TODO nodes/spans/phases/organisations should not be set in network - check this and warn if so

                    # Is data already seen?
                    if network_id in self._networks:
                        # Is it inconsistent with what we have seen before?
                        if json_diff_function(
                            self._networks[network_id]["network_data_original"], network
                        ):
                            # record error for later
                            self._inconsistent_network_ids_seen.add(network_id)

                            # Check references to phases in contracts are consistent with what we have seen before.
                            # This will give a more specific error on the phase_id.
                            # (However, 2 errors will be recorded - one for network_id and one for phase_id!
                            #  Can we make that tidier?
                            #  Suspect as we work to make errors more informative that will happen anyway #TODO)
                            if "contracts" in self._networks[network_id][
                                "network_data_output"
                            ] and isinstance(
                                self._networks[network_id]["network_data_output"][
                                    "contracts"
                                ],
                                list,
                            ):
                                for contract in self._networks[network_id][
                                    "network_data_output"
                                ]["contracts"]:
                                    if "relatedPhases" in contract and isinstance(
                                        contract["relatedPhases"], list
                                    ):
                                        for phase_reference in contract[
                                            "relatedPhases"
                                        ]:
                                            self._process_phase(
                                                network_id, phase_reference
                                            )
                    else:

                        # Not seen this before!

                        # Store it
                        self._networks[network_id] = {
                            "network_data_original": copy.deepcopy(network),
                            "network_data_output": copy.deepcopy(network),
                            "nodes": [],
                            "spans": [],
                            "phases": {},
                            "organisations": {},
                        }

                        # Sort references to phases in contracts
                        # (Must do after storing, as this writes data to the stored network)
                        if "contracts" in self._networks[network_id][
                            "network_data_output"
                        ] and isinstance(
                            self._networks[network_id]["network_data_output"][
                                "contracts"
                            ],
                            list,
                        ):
                            for contract in self._networks[network_id][
                                "network_data_output"
                            ]["contracts"]:
                                if "relatedPhases" in contract and isinstance(
                                    contract["relatedPhases"], list
                                ):
                                    out: list = []
                                    for phase_reference in contract["relatedPhases"]:
                                        phase_data = self._process_phase(
                                            network_id, phase_reference
                                        )
                                        if phase_data:
                                            out.append(phase_data)
                                        else:
                                            out.append(phase_reference)
                                    contract["relatedPhases"] = out

    def _process_node(self, geojson_feature_node: dict) -> None:
        node = copy.deepcopy(geojson_feature_node.get("properties", {}))
        for key_to_remove in ["network", "featureType"]:
            if key_to_remove in node:
                del node[key_to_remove]
        network_id = (
            geojson_feature_node.get("properties", {}).get("network", {}).get("id")
        )
        if network_id not in self._networks.keys():
            # TODO log error
            return

        # sort organisations
        if isinstance(node.get("physicalInfrastructureProvider"), dict):
            node["physicalInfrastructureProvider"] = self._process_organisation(
                network_id, node["physicalInfrastructureProvider"]
            )
        if isinstance(node.get("networkProviders"), list):
            node["networkProviders"] = [
                self._process_organisation(network_id, i)
                for i in node["networkProviders"]
                if isinstance(i, dict)
            ]

        # sort phase
        if isinstance(node.get("phase"), dict) and node.get("phase"):
            phase_data = self._process_phase(network_id, node["phase"])
            if phase_data:
                node["phase"] = phase_data

        if geojson_feature_node.get("geometry"):
            node["location"] = geojson_feature_node["geometry"]

        self._networks[network_id]["nodes"].append(node)

    def _process_span(self, geojson_feature_span: dict) -> None:
        span = copy.deepcopy(geojson_feature_span.get("properties", {}))
        for key_to_remove in ["network", "featureType"]:
            if key_to_remove in span:
                del span[key_to_remove]
        network_id = (
            geojson_feature_span.get("properties", {}).get("network", {}).get("id")
        )
        if network_id not in self._networks.keys():
            # TODO log error
            return

        # sort organisations
        if isinstance(span.get("physicalInfrastructureProvider"), dict):
            span["physicalInfrastructureProvider"] = self._process_organisation(
                network_id, span["physicalInfrastructureProvider"]
            )
        if isinstance(span.get("networkProviders"), list):
            span["networkProviders"] = [
                self._process_organisation(network_id, i)
                for i in span["networkProviders"]
                if isinstance(i, dict)
            ]

        # sort phase
        if isinstance(span.get("phase"), dict) and span.get("phase"):
            phase_data = self._process_phase(network_id, span["phase"])
            if phase_data:
                span["phase"] = phase_data

        if geojson_feature_span.get("geometry"):
            span["route"] = geojson_feature_span["geometry"]

        if isinstance(span.get("start"), dict):
            span["start"] = span["start"].get("id")
        if isinstance(span.get("end"), dict):
            span["end"] = span["end"].get("id")

        self._networks[network_id]["spans"].append(span)

    def _process_phase(self, network_id: str, phase: dict) -> Optional[dict]:
        phase_id = phase.get("id")
        # If no id, can't do anything. TODO log somewhere?
        if not phase_id or not isinstance(phase_id, str):
            return None
        # Check funders
        funders = phase.get("funders")
        if isinstance(funders, list) and funders:
            new_funders = []
            for funder in funders:
                funder_data = self._process_organisation(network_id, funder)
                if funder_data:
                    new_funders.append(funder_data)
            phase["funders"] = new_funders
        # Check data
        if phase_id in self._networks[network_id]["phases"]:
            # Is it inconsistent with what we have seen before?
            if json_diff_function(
                self._networks[network_id]["phases"][phase_id], phase
            ):
                self._inconsistent_phase_ids_by_network_id[network_id].add(phase_id)
        else:
            # Not seen this before; store it
            self._networks[network_id]["phases"][phase_id] = phase
        # Make output
        out: dict = {"id": phase_id}
        # Take name from data on network, not data that is passed to this function.
        # This means that if inconsistent names are in input, we'll have consistent names in the output.
        name = self._networks[network_id]["phases"][phase_id].get("name")
        if name:
            out["name"] = name
        return out

    def _process_organisation(self, network_id: str, organisation: dict) -> dict:
        organisation_id = organisation.get("id")
        # If no id, can't do anything. TODO log somewhere?
        if not organisation_id or not isinstance(organisation_id, str):
            return organisation
        # Check data
        if organisation_id in self._networks[network_id]["organisations"]:
            # Is it inconsistent with what we have seen before?
            if json_diff_function(
                self._networks[network_id]["organisations"][organisation_id],
                organisation,
            ):
                self._inconsistent_organisation_ids_by_network_id[network_id].add(
                    organisation_id
                )
        else:
            # Not seen this before; store it
            self._networks[network_id]["organisations"][organisation_id] = organisation
        # Make output
        out: dict = {"id": organisation_id}
        # Take name from data on network, not data that is passed to this function.
        # This means that if inconsistent names are in input, we'll have consistent names in the output.
        name = self._networks[network_id]["organisations"][organisation_id].get("name")
        if name:
            out["name"] = name
        return out

    def get_json(self) -> dict:
        """After processing, call to get JSON output."""
        out: dict = {"networks": []}
        for network in self._networks.values():
            # We are going to change network, so we need to take a copy
            network_data = copy.deepcopy(network["network_data_output"])
            # Copy other data we have built to network_data
            # Arrays have minItems: 1 set - so only add if we actually have content
            for key in ["nodes", "spans"]:
                if network[key]:
                    network_data[key] = network[key]
            for key in ["phases", "organisations"]:
                if network[key]:
                    network_data[key] = list(network[key].values())
            # build return
            out["networks"].append(network_data)
        return out

    def get_meta_json(self) -> dict:
        """After processing, call to get meta information on conversion."""
        out: dict = {"output_field_coverage": {}}
        # field coverage
        for key, value in fields_present_generator(self.get_json()):
            if key not in out["output_field_coverage"]:
                out["output_field_coverage"][key] = {"count": 1}
            else:
                out["output_field_coverage"][key]["count"] += 1
        # inconsistent
        out["inconsistent_phases_by_network_id"] = {
            k: {"phase_ids": sorted(list(v))}
            for k, v in self._inconsistent_phase_ids_by_network_id.items()
        }
        out["inconsistent_organisations_by_network_id"] = {
            k: {"organisation_ids": sorted(list(v))}
            for k, v in self._inconsistent_organisation_ids_by_network_id.items()
        }
        out["inconsistent_network_ids"] = list(self._inconsistent_network_ids_seen)
        # return
        return out



import argparse
import json

import libcoveofds.additionalfields
import libcoveofds.geojson
import libcoveofds.jsonschemavalidate
import libcoveofds.python_validate
import libcoveofds.schema


def main():
    parser = argparse.ArgumentParser(description="Lib Cove OFDS CLI")
    subparsers = parser.add_subparsers(dest="subparser_name")

    python_validate_parser = subparsers.add_parser(
        "pythonvalidate",
        aliases=["pv"],
        help="Check that data conforms to normative rules specified in OFDS",
    )
    python_validate_parser.add_argument(
        "inputfilename", help="File name of an input JSON data file"
    )

    additional_fields_parser = subparsers.add_parser(
        "additionalfields",
        aliases=["af"],
        help="Report additional fields not specified in the schema",
    )
    additional_fields_parser.add_argument(
        "inputfilename", help="File name of an input JSON data file"
    )

    json_schema_validate_parser = subparsers.add_parser(
        "jsonschemavalidate", aliases=["jsv"], help="Validate data against the schema"
    )
    json_schema_validate_parser.add_argument(
        "inputfilename", help="File name of an input JSON data file"
    )

    json_to_geojson_parser = subparsers.add_parser(
        "jsontogeojson",
        aliases=["jtogj"],
        help="Convert data from JSON format to GeoJSON format",
    )
    json_to_geojson_parser.add_argument(
        "inputfilename", help="File name of an input JSON data file"
    )
    json_to_geojson_parser.add_argument(
        "outputnodesfilename", help="Output filename to write Nodes GeoJSON data to"
    )
    json_to_geojson_parser.add_argument(
        "outputspansfilename", help="Output filename to write Spans GeoJSON data to"
    )
    json_to_geojson_parser.add_argument(
        "--outputmetafilename",
        help="Output filename to write meta JSON data to",
        required=False,
    )

    geojson_to_json_parser = subparsers.add_parser(
        "geojsontojson",
        aliases=["gjtoj"],
        help="Convert data from GeoJSON to JSON format",
    )
    geojson_to_json_parser.add_argument(
        "inputnodesfilename", help="File name of an input Nodes GeoJSON data file"
    )
    geojson_to_json_parser.add_argument(
        "inputspansfilename", help="File name of an input Spans GeoJSON data file"
    )
    geojson_to_json_parser.add_argument(
        "outputfilename", help="Output filename to write JSON data to"
    )
    geojson_to_json_parser.add_argument(
        "--outputmetafilename",
        help="Output filename to write meta JSON data to",
        required=False,
    )

    args = parser.parse_args()

    if args.subparser_name == "pythonvalidate" or args.subparser_name == "pv":

        with open(args.inputfilename) as fp:
            input_data = json.load(fp)

        schema = libcoveofds.schema.OFDSSchema()
        validator = libcoveofds.python_validate.PythonValidate(schema)

        output = validator.validate(input_data)

        print(json.dumps(output, indent=4))

    elif args.subparser_name == "additionalfields" or args.subparser_name == "af":

        with open(args.inputfilename) as fp:
            input_data = json.load(fp)

        schema = libcoveofds.schema.OFDSSchema()
        validator = libcoveofds.additionalfields.AdditionalFields(schema)

        output = validator.process(input_data)

        print(json.dumps(output, indent=4))

    elif args.subparser_name == "jsonschemavalidate" or args.subparser_name == "jsv":

        with open(args.inputfilename) as fp:
            input_data = json.load(fp)

        schema = libcoveofds.schema.OFDSSchema()
        validators = libcoveofds.jsonschemavalidate.JSONSchemaValidator(schema)

        output = validators.validate(input_data)

        output_json = [o.json() for o in output]

        print(json.dumps(output_json, indent=4))

    if args.subparser_name == "jsontogeojson" or args.subparser_name == "jtogj":

        with open(args.inputfilename) as fp:
            input_data = json.load(fp)

        converter = libcoveofds.geojson.JSONToGeoJSONConverter()
        converter.process_package(input_data)

        with open(args.outputnodesfilename, "w") as fp:
            json.dump(converter.get_nodes_geojson(), fp, indent=4)

        with open(args.outputspansfilename, "w") as fp:
            json.dump(converter.get_spans_geojson(), fp, indent=4)

        if args.outputmetafilename:
            with open(args.outputmetafilename, "w") as fp:
                json.dump(converter.get_meta_json(), fp, indent=4)

    elif args.subparser_name == "geojsontojson" or args.subparser_name == "gjtoj":

        with open(args.inputnodesfilename) as fp:
            nodes_data = json.load(fp)

        with open(args.inputspansfilename) as fp:
            spans_data = json.load(fp)

        converter = libcoveofds.geojson.GeoJSONToJSONConverter()
        converter.process_data(
            nodes_data,
            assumed_feature_type=libcoveofds.geojson.GeoJSONAssumeFeatureType.NODE,
        )
        converter.process_data(
            spans_data,
            assumed_feature_type=libcoveofds.geojson.GeoJSONAssumeFeatureType.SPAN,
        )

        with open(args.outputfilename, "w") as fp:
            json.dump(converter.get_json(), fp, indent=4)

        if args.outputmetafilename:
            with open(args.outputmetafilename, "w") as fp:
                json.dump(converter.get_meta_json(), fp, indent=4)


if __name__ == "__main__":
    main()



from collections import defaultdict

from libcoveofds.schema import OFDSSchema


class AdditionalCheckForNetwork:
    """Any check that wants to be provided should extend this abstract class and overwrite methods"""

    def __init__(self, schema_object: OFDSSchema):
        self._additional_check_results: list = []
        self._schema_object: OFDSSchema = schema_object

    def check_node_first_pass(self, node: dict, path: str):
        pass

    def check_span_first_pass(self, span: dict, path: str):
        pass

    def check_phase_first_pass(self, phase: dict, path: str):
        pass

    def check_organisation_first_pass(self, organisation: dict, path: str):
        pass

    def check_contract_first_pass(self, contract: dict, path: str):
        pass

    def check_node_second_pass(self, node: dict, path: str):
        pass

    def check_span_second_pass(self, span: dict, path: str):
        pass

    def check_phase_second_pass(self, phase: dict, path: str):
        pass

    def check_organisation_second_pass(self, organisation: dict, path: str):
        pass

    def check_contract_second_pass(self, contract: dict, path: str):
        pass

    def get_additional_check_results(self) -> list:
        return self._additional_check_results

    def skip_if_any_links_have_external_node_data(self) -> bool:
        return False

    def skip_if_any_links_have_external_span_data(self) -> bool:
        return False


class SpansMustHaveValidNodesAdditionalCheckForNetwork(AdditionalCheckForNetwork):
    def __init__(self, schema_object: OFDSSchema):
        super().__init__(schema_object)
        self._node_ids_seen: list = []

    def check_node_first_pass(self, node: dict, path: str):
        id = node.get("id")
        if id:
            self._node_ids_seen.append(id)

    def check_span_second_pass(self, span: dict, path: str):
        span_id = span.get("id")
        start = span.get("start")
        if start and isinstance(start, str) and not start in self._node_ids_seen:
            self._additional_check_results.append(
                {
                    "type": "span_start_node_not_found",
                    "missing_node_id": start,
                    "span_id": span_id,
                    "path": path + "/start",
                }
            )
        end = span.get("end")
        if end and isinstance(end, str) and not end in self._node_ids_seen:
            self._additional_check_results.append(
                {
                    "type": "span_end_node_not_found",
                    "missing_node_id": end,
                    "span_id": span_id,
                    "path": path + "/end",
                }
            )

    def skip_if_any_links_have_external_node_data(self) -> bool:
        return True

    def skip_if_any_links_have_external_span_data(self) -> bool:
        return True


class NodesLocationAndSpansRouteAdditionalCheckForNetwork(AdditionalCheckForNetwork):
    def check_node_first_pass(self, node: dict, path: str):
        location = node.get("location")
        if isinstance(location, dict) and location:
            type = location.get("type")
            if type != "Point":
                self._additional_check_results.append(
                    {
                        "type": "node_location_type_incorrect",
                        "node_id": node.get("id"),
                        "incorrect_type": type,
                        "path": path + "/location/type",
                    }
                )
            if not self._is_json_coordinates(location.get("coordinates")):
                self._additional_check_results.append(
                    {
                        "type": "node_location_coordinates_incorrect",
                        "node_id": node.get("id"),
                        "incorrect_coordinates": location.get("coordinates"),
                        "path": path + "/location/coordinates",
                    }
                )

    def check_span_first_pass(self, span: dict, path: str):
        location = span.get("route")
        if isinstance(location, dict) and location:
            type = location.get("type")
            if type != "LineString":
                self._additional_check_results.append(
                    {
                        "type": "span_route_type_incorrect",
                        "span_id": span.get("id"),
                        "incorrect_type": type,
                        "path": path + "/route/type",
                    }
                )
            if not self._is_json_list_coordinates(location.get("coordinates")):
                self._additional_check_results.append(
                    {
                        "type": "span_route_coordinates_incorrect",
                        "span_id": span.get("id"),
                        "incorrect_coordinates": location.get("coordinates"),
                        "path": path + "/route/coordinates",
                    }
                )

    def _is_json_list_coordinates(self, list_coordinates):
        if not isinstance(list_coordinates, list):
            return False
        for coordinates in list_coordinates:
            if not self._is_json_coordinates(coordinates):
                return False
        return True

    def _is_json_coordinates(self, coordinates):
        return (
            isinstance(coordinates, list)
            and len(coordinates) == 2
            and (isinstance(coordinates[0], float) or isinstance(coordinates[0], int))
            and (isinstance(coordinates[1], float) or isinstance(coordinates[1], int))
        )


class PhaseReferenceAdditionalCheckForNetwork(AdditionalCheckForNetwork):
    def __init__(self, schema_object: OFDSSchema):
        super().__init__(schema_object)
        self._phases: dict = {}

    def check_phase_first_pass(self, phase: dict, path: str):
        id = phase.get("id")
        name = phase.get("name")
        if id:
            self._phases[id] = name

    def check_node_second_pass(self, node: dict, path: str):
        if "phase" in node and isinstance(node["phase"], dict):
            self._check_related_phase_object(
                node["phase"],
                {
                    "type": "node_phase_reference_id_not_found",
                    "node_id": node.get("id"),
                    "path": path + "/phase/id",
                },
                {
                    "type": "node_phase_reference_name_does_not_match",
                    "node_id": node.get("id"),
                    "path": path + "/phase/name",
                },
                {
                    "type": "node_phase_reference_name_set_but_not_in_original",
                    "node_id": node.get("id"),
                    "path": path + "/phase/name",
                },
            )

    def check_span_second_pass(self, span: dict, path: str):
        if "phase" in span and isinstance(span["phase"], dict):
            self._check_related_phase_object(
                span["phase"],
                {
                    "type": "span_phase_reference_id_not_found",
                    "span_id": span.get("id"),
                    "path": path + "/phase/id",
                },
                {
                    "type": "span_phase_reference_name_does_not_match",
                    "span_id": span.get("id"),
                    "path": path + "/phase/name",
                },
                {
                    "type": "span_phase_reference_name_set_but_not_in_original",
                    "span_id": span.get("id"),
                    "path": path + "/phase/name",
                },
            )

    def check_contract_second_pass(self, contract: dict, path: str):
        if "relatedPhases" in contract and isinstance(contract["relatedPhases"], list):
            for related_phase_idx, related_phase in enumerate(
                contract["relatedPhases"]
            ):
                if isinstance(related_phase, dict):
                    self._check_related_phase_object(
                        related_phase,
                        {
                            "type": "contract_related_phase_reference_id_not_found",
                            "contract_id": contract.get("id"),
                            "path": path
                            + "/relatedPhases/"
                            + str(related_phase_idx)
                            + "/id",
                        },
                        {
                            "type": "contract_related_phase_reference_name_does_not_match",
                            "contract_id": contract.get("id"),
                            "path": path
                            + "/relatedPhases/"
                            + str(related_phase_idx)
                            + "/name",
                        },
                        {
                            "type": "contract_related_phase_reference_name_set_but_not_in_original",
                            "contract_id": contract.get("id"),
                            "path": path
                            + "/relatedPhases/"
                            + str(related_phase_idx)
                            + "/name",
                        },
                    )

    def _check_related_phase_object(
        self,
        related_phase: dict,
        id_not_found_result: dict,
        name_not_match_result: dict,
        name_set_but_not_in_original_result: dict,
    ):
        id = related_phase.get("id")
        name = related_phase.get("name")
        # id is required in JSON Schema - if it is not set we can let that validation raise an error.
        # We'll only carry on with our checks (those that can't be done in JSON Schema) if id exists.
        if id:
            if id in self._phases:
                # check - if name is set on reference but not on original
                if name and not self._phases[id]:
                    name_set_but_not_in_original_result.update(
                        {"name_in_reference": name}
                    )
                    self._additional_check_results.append(
                        name_set_but_not_in_original_result
                    )
                # check - if names are both set, do they match?
                if name and self._phases[id] and name != self._phases[id]:
                    name_not_match_result.update(
                        {"name_in_reference": name, "name_should_be": self._phases[id]}
                    )
                    self._additional_check_results.append(name_not_match_result)
            else:
                # check failed - id is not known
                id_not_found_result.update({"phase_id_not_found": id})
                self._additional_check_results.append(id_not_found_result)


class OrganisationReferenceAdditionalCheckForNetwork(AdditionalCheckForNetwork):
    def __init__(self, schema_object: OFDSSchema):
        super().__init__(schema_object)
        self._organisations: dict = {}

    def check_organisation_first_pass(self, organisation: dict, path: str):
        id = organisation.get("id")
        name = organisation.get("name")
        if id:
            self._organisations[id] = name

    def check_node_second_pass(self, node: dict, path: str):
        if "physicalInfrastructureProvider" in node and isinstance(
            node["physicalInfrastructureProvider"], dict
        ):
            self._check_related_organisation_object(
                node["physicalInfrastructureProvider"],
                {
                    "type": "node_organisation_reference_id_not_found",
                    "node_id": node.get("id"),
                    "field": "physicalInfrastructureProvider",
                    "path": path + "/physicalInfrastructureProvider/id",
                },
                {
                    "type": "node_organisation_reference_name_does_not_match",
                    "node_id": node.get("id"),
                    "field": "physicalInfrastructureProvider",
                    "path": path + "/physicalInfrastructureProvider/name",
                },
                {
                    "type": "node_organisation_reference_name_set_but_not_in_original",
                    "node_id": node.get("id"),
                    "field": "physicalInfrastructureProvider",
                    "path": path + "/physicalInfrastructureProvider/name",
                },
            )
        if "networkProviders" in node and isinstance(node["networkProviders"], list):
            for network_provider_idx, network_provider in enumerate(
                node["networkProviders"]
            ):
                if isinstance(network_provider, dict):
                    self._check_related_organisation_object(
                        network_provider,
                        {
                            "type": "node_organisation_reference_id_not_found",
                            "node_id": node.get("id"),
                            "field": "networkProviders",
                            "path": path
                            + "/networkProviders/"
                            + str(network_provider_idx)
                            + "/id",
                        },
                        {
                            "type": "node_organisation_reference_name_does_not_match",
                            "node_id": node.get("id"),
                            "field": "networkProviders",
                            "path": path
                            + "/networkProviders/"
                            + str(network_provider_idx)
                            + "/name",
                        },
                        {
                            "type": "node_organisation_reference_name_set_but_not_in_original",
                            "node_id": node.get("id"),
                            "field": "networkProviders",
                            "path": path
                            + "/networkProviders/"
                            + str(network_provider_idx)
                            + "/name",
                        },
                    )

    def check_span_second_pass(self, span: dict, path: str):
        if "physicalInfrastructureProvider" in span and isinstance(
            span["physicalInfrastructureProvider"], dict
        ):
            self._check_related_organisation_object(
                span["physicalInfrastructureProvider"],
                {
                    "type": "span_organisation_reference_id_not_found",
                    "span_id": span.get("id"),
                    "field": "physicalInfrastructureProvider",
                    "path": path + "/physicalInfrastructureProvider/id",
                },
                {
                    "type": "span_organisation_reference_name_does_not_match",
                    "span_id": span.get("id"),
                    "field": "physicalInfrastructureProvider",
                    "path": path + "/physicalInfrastructureProvider/name",
                },
                {
                    "type": "span_organisation_reference_name_set_but_not_in_original",
                    "span_id": span.get("id"),
                    "field": "physicalInfrastructureProvider",
                    "path": path + "/physicalInfrastructureProvider/name",
                },
            )
        if "networkProviders" in span and isinstance(span["networkProviders"], list):
            for network_provider_idx, network_provider in enumerate(
                span["networkProviders"]
            ):
                if isinstance(network_provider, dict):
                    self._check_related_organisation_object(
                        network_provider,
                        {
                            "type": "span_organisation_reference_id_not_found",
                            "span_id": span.get("id"),
                            "field": "networkProviders",
                            "path": path
                            + "/networkProviders/"
                            + str(network_provider_idx)
                            + "/id",
                        },
                        {
                            "type": "span_organisation_reference_name_does_not_match",
                            "span_id": span.get("id"),
                            "field": "networkProviders",
                            "path": path
                            + "/networkProviders/"
                            + str(network_provider_idx)
                            + "/name",
                        },
                        {
                            "type": "span_organisation_reference_name_set_but_not_in_original",
                            "span_id": span.get("id"),
                            "field": "networkProviders",
                            "path": path
                            + "/networkProviders/"
                            + str(network_provider_idx)
                            + "/name",
                        },
                    )
        if "supplier" in span and isinstance(span["supplier"], dict):
            self._check_related_organisation_object(
                span["supplier"],
                {
                    "type": "span_organisation_reference_id_not_found",
                    "span_id": span.get("id"),
                    "field": "supplier",
                    "path": path + "/supplier/id",
                },
                {
                    "type": "span_organisation_reference_name_does_not_match",
                    "span_id": span.get("id"),
                    "field": "supplier",
                    "path": path + "/supplier/name",
                },
                {
                    "type": "span_organisation_reference_name_set_but_not_in_original",
                    "span_id": span.get("id"),
                    "field": "supplier",
                    "path": path + "/supplier/name",
                },
            )

    def check_phase_second_pass(self, phase: dict, path: str):
        if "funders" in phase and isinstance(phase["funders"], list):
            for funder_idx, funder in enumerate(phase["funders"]):
                if isinstance(funder, dict):
                    self._check_related_organisation_object(
                        funder,
                        {
                            "type": "phase_organisation_reference_id_not_found",
                            "phase_id": phase.get("id"),
                            "path": path + "/funders/" + str(funder_idx) + "/id",
                        },
                        {
                            "type": "phase_organisation_reference_name_does_not_match",
                            "phase_id": phase.get("id"),
                            "path": path + "/funders/" + str(funder_idx) + "/name",
                        },
                        {
                            "type": "phase_organisation_reference_name_set_but_not_in_original",
                            "phase_id": phase.get("id"),
                            "path": path + "/funders/" + str(funder_idx) + "/name",
                        },
                    )

    def _check_related_organisation_object(
        self,
        related_organisation: dict,
        id_not_found_result: dict,
        name_not_match_result: dict,
        name_set_but_not_in_original_result: dict,
    ):
        id = related_organisation.get("id")
        name = related_organisation.get("name")
        # id is required in JSON Schema - if it is not set we can let that validation raise an error.
        # We'll only carry on with our checks (those that can't be done in JSON Schema) if id exists.
        if id:
            if id in self._organisations:
                # check - if name is set on reference but not on original
                if name and not self._organisations[id]:
                    name_set_but_not_in_original_result.update(
                        {"name_in_reference": name}
                    )
                    self._additional_check_results.append(
                        name_set_but_not_in_original_result
                    )
                # check - if names are both set, do they match?
                if name and self._organisations[id] and name != self._organisations[id]:
                    name_not_match_result.update(
                        {
                            "name_in_reference": name,
                            "name_should_be": self._organisations[id],
                        }
                    )
                    self._additional_check_results.append(name_not_match_result)
            else:
                # check failed - id is not known
                id_not_found_result.update({"organisation_id_not_found": id})
                self._additional_check_results.append(id_not_found_result)


class NodeInternationalConnectionCountryAdditionalCheckForNetwork(
    AdditionalCheckForNetwork
):
    def check_node_first_pass(self, node: dict, path: str):
        if "internationalConnections" in node and isinstance(
            node["internationalConnections"], list
        ):
            for international_connection_idx, international_connection in enumerate(
                node["internationalConnections"]
            ):
                if isinstance(
                    international_connection, dict
                ) and not international_connection.get("country"):
                    self._additional_check_results.append(
                        {
                            "type": "node_international_connections_country_not_set",
                            "node_id": node.get("id"),
                            "path": path
                            + "/internationalConnections/"
                            + str(international_connection_idx),
                        }
                    )


class IsNodeUsedInSpanAdditionalCheckForNetwork(AdditionalCheckForNetwork):
    def __init__(self, schema_object: OFDSSchema):
        super().__init__(schema_object)
        self._node_ids_used_in_spans: list = []

    def check_span_first_pass(self, span: dict, path: str):
        start = span.get("start")
        if start and start not in self._node_ids_used_in_spans:
            self._node_ids_used_in_spans.append(start)
        end = span.get("end")
        if end and end not in self._node_ids_used_in_spans:
            self._node_ids_used_in_spans.append(end)

    def check_node_second_pass(self, node: dict, path: str):
        id = node.get("id")
        if id and id not in self._node_ids_used_in_spans:
            self._additional_check_results.append(
                {
                    "type": "node_not_used_in_any_spans",
                    "node_id": node.get("id"),
                    "path": path,
                }
            )

    def skip_if_any_links_have_external_node_data(self) -> bool:
        return True

    def skip_if_any_links_have_external_span_data(self) -> bool:
        return True


class UniqueIDsAdditionalCheckForNetwork(AdditionalCheckForNetwork):
    def __init__(self, schema_object: OFDSSchema):
        super().__init__(schema_object)
        self._node_ids_seen: defaultdict = defaultdict(list)
        self._span_ids_seen: defaultdict = defaultdict(list)
        self._phase_ids_seen: defaultdict = defaultdict(list)
        self._organisation_ids_seen: defaultdict = defaultdict(list)
        self._contract_ids_seen: defaultdict = defaultdict(list)

    def check_node_first_pass(self, node: dict, path: str):
        id = node.get("id")
        if id and isinstance(id, str):
            self._node_ids_seen[id].append(path)

    def check_span_first_pass(self, span: dict, path: str):
        id = span.get("id")
        if id and isinstance(id, str):
            self._span_ids_seen[id].append(path)

    def check_phase_first_pass(self, phase: dict, path: str):
        id = phase.get("id")
        if id and isinstance(id, str):
            self._phase_ids_seen[id].append(path)

    def check_organisation_first_pass(self, organisation: dict, path: str):
        id = organisation.get("id")
        if id and isinstance(id, str):
            self._organisation_ids_seen[id].append(path)

    def check_contract_first_pass(self, contract: dict, path: str):
        id = contract.get("id")
        if id and isinstance(id, str):
            self._contract_ids_seen[id].append(path)

    def get_additional_check_results(self) -> list:
        out: list = []
        for id, paths in self._node_ids_seen.items():
            if len(paths) > 1:
                for path in paths:
                    out.append(
                        {
                            "type": "duplicate_node_id",
                            "node_id": id,
                            "path": path + "/id",
                        }
                    )
        for id, paths in self._span_ids_seen.items():
            if len(paths) > 1:
                for path in paths:
                    out.append(
                        {
                            "type": "duplicate_span_id",
                            "span_id": id,
                            "path": path + "/id",
                        }
                    )
        for id, paths in self._phase_ids_seen.items():
            if len(paths) > 1:
                for path in paths:
                    out.append(
                        {
                            "type": "duplicate_phase_id",
                            "phase_id": id,
                            "path": path + "/id",
                        }
                    )
        for id, paths in self._organisation_ids_seen.items():
            if len(paths) > 1:
                for path in paths:
                    out.append(
                        {
                            "type": "duplicate_organisation_id",
                            "organisation_id": id,
                            "path": path + "/id",
                        }
                    )
        for id, paths in self._contract_ids_seen.items():
            if len(paths) > 1:
                for path in paths:
                    out.append(
                        {
                            "type": "duplicate_contract_id",
                            "contract_id": id,
                            "path": path + "/id",
                        }
                    )
        return out

    def skip_if_any_links_have_external_node_data(self) -> bool:
        return False

    def skip_if_any_links_have_external_span_data(self) -> bool:
        return False


ADDITIONAL_CHECK_CLASSES_FOR_NETWORK = [
    SpansMustHaveValidNodesAdditionalCheckForNetwork,
    NodesLocationAndSpansRouteAdditionalCheckForNetwork,
    PhaseReferenceAdditionalCheckForNetwork,
    OrganisationReferenceAdditionalCheckForNetwork,
    NodeInternationalConnectionCountryAdditionalCheckForNetwork,
    IsNodeUsedInSpanAdditionalCheckForNetwork,
    UniqueIDsAdditionalCheckForNetwork,
]


class PythonValidate:
    """Validates data using additional checks custom written in Python"""

    def __init__(self, schema: OFDSSchema):
        self._schema = schema

    def validate(self, json_data: dict) -> list:
        """Call with data. Results are returned."""

        additional_checks: list = []

        # For each Network
        networks = json_data.get("networks")
        if isinstance(networks, list):
            for network_idx, network in enumerate(networks):
                if isinstance(network, dict):
                    additional_check_instances = [
                        x(self._schema) for x in ADDITIONAL_CHECK_CLASSES_FOR_NETWORK
                    ]
                    links = network.get("links", [])
                    links_with_external_nodes = [
                        l
                        for l in links
                        if isinstance(l, dict)
                        and l.get("rel", "")
                        in self._schema.get_link_rels_for_external_nodes()
                    ]
                    if links_with_external_nodes:
                        additional_check_instances = [
                            x
                            for x in additional_check_instances
                            if not x.skip_if_any_links_have_external_node_data()
                        ]
                        additional_checks.append(
                            {
                                "network_id": network.get("id"),
                                "type": "has_links_with_external_node_data",
                                "path": "/networks/" + str(network_idx),
                            }
                        )
                    links_with_external_spans = [
                        l
                        for l in links
                        if isinstance(l, dict)
                        and l.get("rel", "")
                        in self._schema.get_link_rels_for_external_spans()
                    ]
                    if links_with_external_spans:
                        additional_check_instances = [
                            x
                            for x in additional_check_instances
                            if not x.skip_if_any_links_have_external_span_data()
                        ]
                        additional_checks.append(
                            {
                                "network_id": network.get("id"),
                                "type": "has_links_with_external_span_data",
                                "path": "/networks/" + str(network_idx),
                            }
                        )
                    nodes = network.get("nodes", [])
                    nodes = nodes if isinstance(nodes, list) else []
                    spans = network.get("spans", [])
                    spans = spans if isinstance(spans, list) else []
                    phases = network.get("phases", [])
                    phases = phases if isinstance(phases, list) else []
                    organisations = network.get("organisations", [])
                    organisations = (
                        organisations if isinstance(organisations, list) else []
                    )
                    contracts = network.get("contracts", [])
                    contracts = contracts if isinstance(contracts, list) else []
                    # First pass
                    for additional_check_instance in additional_check_instances:
                        for node_idx, node in enumerate(nodes):
                            additional_check_instance.check_node_first_pass(
                                node,
                                "/networks/"
                                + str(network_idx)
                                + "/nodes/"
                                + str(node_idx),
                            )
                        for span_idx, span in enumerate(spans):
                            additional_check_instance.check_span_first_pass(
                                span,
                                "/networks/"
                                + str(network_idx)
                                + "/spans/"
                                + str(span_idx),
                            )
                        for phase_idx, phase in enumerate(phases):
                            additional_check_instance.check_phase_first_pass(
                                phase,
                                "/networks/"
                                + str(network_idx)
                                + "/phases/"
                                + str(phase_idx),
                            )
                        for organisation_idx, organisation in enumerate(organisations):
                            additional_check_instance.check_organisation_first_pass(
                                organisation,
                                "/networks/"
                                + str(network_idx)
                                + "/organisations/"
                                + str(organisation_idx),
                            )
                        for contract_idx, contract in enumerate(contracts):
                            additional_check_instance.check_contract_first_pass(
                                contract,
                                "/networks/"
                                + str(network_idx)
                                + "/contracts/"
                                + str(contract_idx),
                            )
                    # Second pass
                    for additional_check_instance in additional_check_instances:
                        for node_idx, node in enumerate(nodes):
                            additional_check_instance.check_node_second_pass(
                                node,
                                "/networks/"
                                + str(network_idx)
                                + "/nodes/"
                                + str(node_idx),
                            )
                        for span_idx, span in enumerate(spans):
                            additional_check_instance.check_span_second_pass(
                                span,
                                "/networks/"
                                + str(network_idx)
                                + "/spans/"
                                + str(span_idx),
                            )
                        for phase_idx, phase in enumerate(phases):
                            additional_check_instance.check_phase_second_pass(
                                phase,
                                "/networks/"
                                + str(network_idx)
                                + "/phases/"
                                + str(phase_idx),
                            )
                        for organisation_idx, organisation in enumerate(organisations):
                            additional_check_instance.check_organisation_second_pass(
                                organisation,
                                "/networks/"
                                + str(network_idx)
                                + "/organisations/"
                                + str(organisation_idx),
                            )
                        for contract_idx, contract in enumerate(contracts):
                            additional_check_instance.check_contract_second_pass(
                                contract,
                                "/networks/"
                                + str(network_idx)
                                + "/contracts/"
                                + str(contract_idx),
                            )
                    # Results
                    for additional_check_instance in additional_check_instances:
                        for (
                            additional_check
                        ) in additional_check_instance.get_additional_check_results():
                            additional_check["network_id"] = network.get("id")
                            additional_checks.append(additional_check)

        # Return
        return additional_checks



import json
import os

from libcove2.common import schema_dict_fields_generator

_schema_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")


class OFDSSchema:
    """Represents and provides information about the schema."""

    package_schema_url: str = os.path.join(_schema_folder, "schema-0-3-0.json")

    def get_package_schema(self):
        with open(self.package_schema_url) as fp:
            return json.load(fp)

    def get_link_rels_for_external_nodes(self) -> list:
        return [
            "tag:opentelecomdata.net,2022:nodesAPI",
            "tag:opentelecomdata.net,2022:nodesFile",
        ]

    def get_link_rels_for_external_spans(self) -> list:
        return [
            "tag:opentelecomdata.net,2022:spansAPI",
            "tag:opentelecomdata.net,2022:spansFile",
        ]

    def get_package_schema_fields(self) -> set:
        return set(schema_dict_fields_generator(self.get_package_schema()))

    def extract_data_ids_from_data_and_path(self, data: dict, path: list) -> dict:
        out: dict = {}
        # network_id
        if len(path) >= 2 and path[0] == "networks":
            network_id = data["networks"][path[1]].get("id")
            if network_id:
                out["network_id"] = network_id
        # other ids
        for field in ["node", "span"]:
            if len(path) >= 4 and path[0] == "networks" and path[2] == field + "s":
                id = data["networks"][path[1]][field + "s"][path[3]].get("id")
                if id:
                    out[field + "_id"] = id
        # return
        return out



from libcove2.common import get_additional_fields_info
from libcoveofds.schema import OFDSSchema


class AdditionalFields:
    """Process data and return additional fields information"""

    def __init__(self, schema: OFDSSchema):
        self._schema = schema

    def process(self, json_data: dict) -> list:
        """Process method. Call with data. Results are returned."""

        schema_fields = self._schema.get_package_schema_fields()

        additional_fields = get_additional_fields_info(json_data, schema_fields)

        return additional_fields



from setuptools import find_packages, setup

setup(
    name="libcoveofds",
    version="0.8.0",
    author="Open Data Services",
    author_email="code@opendataservices.coop",
    url="https://github.com/Open-Telecoms-Data/lib-cove-ofds",
    project_urls={
        "Documentation": "https://libcoveofds.readthedocs.io/en/latest/",
        "Issues": "https://github.com/Open-Telecoms-Data/lib-cove-ofds/issues",
        "Source": "https://github.com/Open-Telecoms-Data/lib-cove-ofds",
    },
    description="A data review library",
    packages=find_packages(),
    long_description="A data review library",
    python_requires=">=3.8",
    install_requires=[
        "jsonschema",
        "json-merge-patch",
        "rfc3987",
    ],
    extras_require={
        "dev": ["pytest", "flake8", "black==22.3.0", "isort", "mypy", "Sphinx"]
    },
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
    ],
    entry_points="""[console_scripts]
libcoveofds = libcoveofds.cli:main""",
    include_package_data=True,
)



from jsonschema import FormatChecker
from jsonschema.exceptions import ValidationError as JSONSchemaExceptionsValidationError
from jsonschema.validators import Draft202012Validator

from libcoveofds.schema import OFDSSchema


class JSONSchemaValidator:
    """Validates data using the JSON Schema method"""

    def __init__(self, schema: OFDSSchema):
        self._schema = schema

    def validate(self, json_data: dict) -> list:
        """Call with data. Results are returned."""
        validator = Draft202012Validator(
            schema=self._schema.get_package_schema(), format_checker=FormatChecker()
        )
        output = []
        for e in validator.iter_errors(json_data):
            output.append(ValidationError(e, json_data, self._schema))
        return output


class ValidationError:
    """Any problems found in data are returned as an instance of this class."""

    def __init__(
        self,
        json_schema_exceptions_validation_error: JSONSchemaExceptionsValidationError,
        json_data: dict,
        schema: OFDSSchema,
    ):
        self._message = json_schema_exceptions_validation_error.message
        self._path = json_schema_exceptions_validation_error.path
        self._schema_path = json_schema_exceptions_validation_error.schema_path
        self._validator = json_schema_exceptions_validation_error.validator
        self._validator_value = json_schema_exceptions_validation_error.validator_value
        self._data_ids = schema.extract_data_ids_from_data_and_path(
            json_data, self._path
        )
        self._context = json_schema_exceptions_validation_error.context
        self._instance = json_schema_exceptions_validation_error.instance

    def json(self):
        """Return representation of this error in JSON."""
        return {
            "message": self._message,
            "path": list(self._path),
            "schema_path": list(self._schema_path),
            "validator": self._validator,
            "data_ids": self._data_ids,
            "validator_value": self._validator_value,
            "context": self._context,
            "instance": self._instance,
        }
